/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <limits.h>
#include <ncurses.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <assert.h>
#include <fcntl.h>
#include <ctype.h>
#include <regex.h>

#include "main.h"

#include "browser.h"
#include "curs_lib.h"
#include "curs_main.h"
#include "keymap.h"
#include "lib.h"
#include "misc.h"
#include "options.h"
#include "playlist.h"

/* `n' is number of entries in browser menu */
#define ENTER_MENU_BROWSER(n) \
{ \
    *prevmenu = *menu; \
    *menu = safe_calloc(1, sizeof(struct menu)); \
    (*menu)->offset = (*prevmenu)->offset; \
    (*menu)->pagelen = (*prevmenu)->pagelen; \
    (*menu)->menu = MENU_BROWSER; \
    (*menu)->search = browser_search; \
    (*menu)->oldcurrent = (*menu)->current = (*menu)->top = 0; \
    (*menu)->entries = n; \
    (*menu)->make_entry = browser_make_menu_entry; \
    TURN_ON_FLAG((*menu)->redraw, REDRAW_FULL); \
}

struct playlist_entry *first_entry;

static int write_song_names_to_file(const char *file, bool ask_if_exist);
static int append_playlist(const char *path);

/*
 * Add directory to the playlist.
 */
void interactive_add_dir(struct menu **menu, struct menu **prevmenu, bool * playlist_changed)
{
    char path[PATH_MAX + 1];
    char buf[8];
    int tmp = -1;

    snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");

    get_key_descr_for_cmd(CMD_BROWSE, buf, sizeof(buf));
    if (enter_path(path, sizeof(path), DIRECTORY | MUST_EXIST | HALF_FILLED,
		 "Directory [%s to browse]: ", buf) == NULL) {
	return;			/* aborted */
    } else if (Browse == FALSE) {
	expand_path(path, sizeof(path));
	tmp = add_dir(path);	/* returns number of songs added */
	if (tmp > 0) {
	    msg("Added %d songs.", tmp);
	    (*menu)->entries += tmp;
	    *playlist_changed = TRUE;
	    TURN_ON_FLAG((*menu)->redraw, REDRAW_INDEX);
	}
	return;
    }
    /* The global variable Browse is TRUE if we reach this far
       (enter_path() is the only routine that changes it) */

    tmp = browser_examine_directory(cwd());
    if (tmp < 0)
	return;

    /* enter the browser menu */
    ENTER_MENU_BROWSER(tmp);
}

/*
 * Add all registered filetypes in the directory pointed to by path
 * to the playlist. Do not ask user for any input.
 * Return -1 if path can't be opened and tell user about it.
 * Else return number of songs added to the playlist.
 * If no registered filetypes were found in path, tell user
 * about it (and return 0 of course).
 */
int add_dir(const char *path)
{
    DIR *dp;
    struct dirent *dirp;
    struct playlist_entry *new, *last = NULL;
    int ret;
    size_t pathlen;
    char *p = NULL;

    dp = opendir(path);
    if (dp == NULL) {
	perror_msg("Can't open directory %s", path);
	return -1;
    }
    if (first_entry != NULL)
	/* make last point to the last node */
	for (last = first_entry; last->next; last = last->next)
	    ;

    p = safe_malloc(strlen(path) + 2);
    strcpy(p, path);
    if (p[strlen(p) - 1] != '/')
	strcat(p, "/");
    pathlen = strlen(p);
    ret = 0;
    while ((dirp = readdir(dp))) {
	if (strcmp(dirp->d_name, ".") == 0 || strcmp(dirp->d_name, "..") == 0)
	    continue;		/* we don't need . and .. */
	if (!has_registered_filename_extension(dirp->d_name))
	    continue;
	new = safe_malloc(sizeof(struct playlist_entry));
	new->next = NULL;
	new->prev = NULL;
	new->toggled = FALSE;
	new->path = safe_malloc(pathlen + 1 + strlen(dirp->d_name));
	strcpy(new->path, p);
	strcat(new->path, dirp->d_name);
	new->song = safe_strdup(dirp->d_name);
	ret++;
	if (first_entry == NULL) {
	    first_entry = last = new;
	} else {
	    new->prev = last;
	    last->next = new;
	    last = new;
	}
    }

    if (ret == 0)
	msg("No files with the filename extensions defined in the "
		  "%s variable found.", option_name(OPT_FILENAME_EXTENSIONS));

    closedir(dp);
    safe_free(p);
    return ret;
}

/*
 * Add file to the playlist.
 */
void interactive_add_song(struct menu **menu, struct menu **prevmenu, bool * playlist_changed)
{
    char path[PATH_MAX + 1];
    char buf[8];
    int tmp;

    snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");

    get_key_descr_for_cmd(CMD_BROWSE, buf, sizeof(buf));
    if (enter_path(path, sizeof(path), REGISTERED_EXT | MUST_EXIST | HALF_FILLED,
		   "File [%s to browse]: ", buf) == NULL)
	return;
    else if (Browse == FALSE) {
	expand_path(path, sizeof(path));
	add_song(path);
	(*menu)->entries++;
	*playlist_changed = TRUE;
	TURN_ON_FLAG((*menu)->redraw, REDRAW_INDEX);
	return;
    }
    /* The global variable Browse is TRUE if we reach this far
       (enter_path() is the only routine that changes it) */

    tmp = browser_examine_directory(cwd());
    if (tmp < 0)
	return;

    ENTER_MENU_BROWSER(tmp);
}

/*
 * Add the song pointed to by path to the playlist. It will not
 * be checked if path exists, that should be done before calling
 * this function. No interaction with the user is done.
 */
void add_song(const char *path)
{
    struct playlist_entry *new, *p;
    char *q;

    new = safe_malloc(sizeof(struct playlist_entry));
    new->next = NULL;
    new->toggled = FALSE;
    new->path = safe_strdup(path);
    q = strrchr(path, '/');
    q++;
    new->song = safe_strdup(q);

    if (first_entry == NULL) {
	first_entry = new;
	new->prev = NULL;
    } else {
	/* make p point to the last node */
	for (p = first_entry; p->next; p = p->next)
	    ;
	new->prev = p;
	p->next = new;
    }
}

void interactive_append_playlist(struct menu *menu, bool * playlist_changed)
{
    char path[PATH_MAX + 1];
    int entries;

    snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");

    if (enter_path(path, sizeof(path), REG_FILE | MUST_EXIST | HALF_FILLED,
		   "Playlist to append: ") == NULL)
	return; /* either user aborted, or the path wasn't a regular file or didn't exist */
    /* playlist needs to be absolute (but we don't want to force the user to
       type the absolute path), ./ will be removed by enter_path() */
    if (path[0] != '/') {
	char buf[PATH_MAX];
	strfcpy(buf, path, sizeof(buf));
	snprintf(path, sizeof(path), "%s%s",
		 cwd(),
		 IS_ROOT_PATH(cwd()) ? "" : "/");
	strncat(path + strlen(path), buf, sizeof(path) - strlen(path));
	path[sizeof(path) - 1] = '\0';
    }
    entries = append_playlist(path);
    if (entries == 0)
	return;

    *playlist_changed = TRUE;
    menu->entries += entries;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Append the playlist given by the path parameter to the end of the current playlist.
 * Return number of songs added.
 */
static int append_playlist(const char *path)
{
    struct playlist_entry *last_entry = NULL, *new = NULL, *first_appended_entry = NULL;
    char line[PATH_MAX + 1], *p = NULL;
    int lnum;
    int num_entries;
    bool toggled = FALSE;
    FILE *fp;

    fp = fopen(path, "r");
    if (fp == NULL) {
	perror_msg("Can't open %s", path);
	return 0;
    }
    /* make last_entry point to the last node */
    last_entry = first_entry;
    if (last_entry) {		/* if there are any entries */
	while (last_entry->next)
	    last_entry = last_entry->next;
    }

    num_entries = 0;
    lnum = 0;
    while (fgets(line, sizeof(line), fp) != NULL) {
	if (isspace(line[0]))
	    continue;		/* empty line */

	*(line + strlen(line) - 1) = '\0';
	lnum++;

	if (line[0] != '/' && line[0] != '~') {
	    msg("Error: one or more of the paths aren't absolute.");
	    free_playlist(&first_appended_entry);
	    return 0;
	}
	/* is the song toggled? */
	if (line[strlen(line) - 1] == '*' && whitespace(line[strlen(line) - 2])) {
	    char *cp = line + strlen(line) - 1 - 1;
	    while (whitespace(*cp))
		cp--;
	    *(cp + 1) = '\0';
	    toggled = TRUE;
	} else
	    toggled = FALSE;


	if (!has_registered_filename_extension(line)) {
	    msg("Error: the filenames doesn't correlate with the %s variable.",
		      option_name(OPT_FILENAME_EXTENSIONS));
	    free_playlist(&first_appended_entry);
	    return 0;
	}
	if (line[0] == '~' && expand_path(line, sizeof(line)) == NULL) {
	    msg("Error: can't expand %s.", line);
	    free_playlist(&first_appended_entry);
	    return 0;
	}

	new = safe_malloc(sizeof(struct playlist_entry));
	new->next = NULL;
	new->toggled = toggled;	/* toggled was just set to t/f */
	new->path = safe_strdup(line);
	p = strrchr(line, '/');	/* will always succeed */
	*p = '\0';
	p++;
	new->song = safe_strdup(p);
	num_entries++;

	/* Save a pointer to the first appended entry so we can free
	   all allocated nodes if there's an error in the playlist. */
	if (num_entries == 1)
	    first_appended_entry = new;

	if (last_entry == NULL) {	/* if there are no entries */
	    new->prev = NULL;
	    first_entry = last_entry = new;
	} else {
	    last_entry->next = new;
	    new->prev = last_entry;
	    last_entry = new;
	}
    }

    return num_entries;
}

void free_playlist(struct playlist_entry **first_playlist_entry)
{
    struct playlist_entry *p, *tmp;

    for (p = *first_playlist_entry; p; p = tmp) {
	tmp = p->next;
	safe_free(p->path);
	safe_free(p->song);
	safe_free(p);
    }

    *first_playlist_entry = NULL;
}

void interactive_load_playlist(struct menu *menu, char **playlist, bool * playlist_changed)
{
    char path[PATH_MAX + 1];
    struct playlist_entry *p;
    int entries;		/* temporary saving of number of entries in the new playlist */
    bool err;			/* will be set to true by load_playlist() if an error occured */

    if (*playlist_changed && yes_or_no(1, "Save current playlist?")) {
      snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");
	if (!*playlist) {	/* if we don't have any playlist open */
	    if (enter_path(path, sizeof(path), REG_FILE | HALF_FILLED, "Save playlist to: ")) {
		if (save_playlist(path, ASK_IF_EXISTS) == -1)
		    return;	/* don't load a new playlist if we can't save the current */
	    } else
		return;		/* user aborted saving of current playlist, so return */
	} else {		/* we have some playlist open, so just overwrite it */
	    if (save_playlist(*playlist, OVERWRITE) == -1)
		return;		/* don't load a new playlist if we can't save the current */
	}
    }
    *playlist_changed = FALSE;

    snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");
    if (enter_path(path, sizeof(path), REG_FILE | MUST_EXIST | HALF_FILLED,
		   "Playlist to load: ") == NULL)
	return;			/* either user aborted, or the path wasn't a regular file or didn't exist */
    /* playlist needs to be absolute (but we don't want to force the user to
       type the absolute path), ./ will be removed by enter_path() */
    if (path[0] != '/') {
	char buf[PATH_MAX];
	strfcpy(buf, path, sizeof(buf));
	snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");
	strncat(path + strlen(path), buf, sizeof(path) - strlen(path));
	path[sizeof(path) - 1] = '\0';
    }
    p = load_playlist(path, &entries, &err);
    if (!p && err == TRUE)
	return;			/* couldn't load playlist */
    menu->entries = entries;
    menu->current = 0;
    free_playlist(&first_entry);
    first_entry = p;
    safe_free(*playlist);
    *playlist = safe_strdup(path);
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
}

struct playlist_entry *load_playlist(char *playlist, int *num_entries, bool * err)
{
    struct playlist_entry *first_node = NULL, *cur = NULL, *new = NULL;
    char line[PATH_MAX + 1], *p = NULL;
    int lnum;
    bool toggled = FALSE;
    FILE *fp;

    assert(playlist);

    *num_entries = 0;

    fp = fopen(playlist, "r");
    if (fp == NULL) {
	perror_msg("Can't open %s", playlist);
	*err = TRUE;
	return NULL;
    }
    *err = FALSE;

    lnum = 0;
    while (fgets(line, sizeof(line), fp) != NULL) {
	if (isspace(line[0]))
	    continue;		/* empty line */

	*(line + strlen(line) - 1) = '\0';
	lnum++;

	if (line[0] != '/' && line[0] != '~') {
	    msg("Error: one or more of the paths aren't absolute.");
	    free_playlist(&first_node);
	    *err = TRUE;
	    return NULL;
	}
	if (line[strlen(line) - 1] == '*' && whitespace(line[strlen(line) - 2])) {
	    char *cp = line + strlen(line) - 1 - 1;
	    while (whitespace(*cp))
		cp--;
	    *(cp + 1) = '\0';
	    toggled = TRUE;
	} else
	    toggled = FALSE;

	if (!has_registered_filename_extension(line)) {
	    msg("Error: the filenames doesn't correlate with the %s variable.",
		      option_name(OPT_FILENAME_EXTENSIONS));
	    free_playlist(&first_node);
	    *err = TRUE;
	    return NULL;
	}
	if (line[0] == '~' && expand_path(line, sizeof(line)) == NULL) {
	    msg("Error: can't expand %s.", line);
	    free_playlist(&first_node);
	    *err = TRUE;
	    return NULL;
	}

	new = safe_malloc(sizeof(struct playlist_entry));
	new->next = NULL;
	new->prev = NULL;
	new->toggled = toggled;	/* toggled was just set to t/f */
	new->path = safe_strdup(line);
	p = strrchr(line, '/');	/* will always succeed */
	*p = '\0';
	p++;
	new->song = safe_strdup(p);
	(*num_entries)++;

	if (first_node == NULL) {
	    first_node = cur = new;
	} else {
	    cur->next = new;
	    new->prev = cur;
	    cur = cur->next;
	}
    }

    return first_node;		/* can be NULL if no song paths found in the playlist file */
}

/*
 * Move the srcindexth entry to the dstindexth position in the playlist.
 * The playlist must not be empty.
 */
void move_entry(int dstindex, int srcindex)
{
    struct playlist_entry *src, *dst;
    int i;

    if (dstindex == srcindex)
	return;

    assert(first_entry);

    i = 0;
    src = first_entry;
    while (i++ < srcindex)
	src = src->next;
    assert(src);

    i = 0;
    dst = first_entry;
    while (i++ < dstindex)
	dst = dst->next;
    assert(dst);

    /* remove src from the linked list */
    if (src->prev == NULL) {	/* if src is the first node in the list */
	src->next->prev = NULL;	/* src->next is never NULL in this case */
	first_entry = src->next;
    } else {			/* src is not the first node in the list */
	src->prev->next = src->next;	/* src->prev is never NULL in this case */
	if (src->next)		/* if src is not the last node in the list */
	    src->next->prev = src->prev;
    }

    /* if the entry is being moved up, */
    /* then insert src just before the node dst points to */
    if (srcindex > dstindex) {
	src->prev = dst->prev;
	if (dst->prev == NULL)	/* if dst is the first node in the list */
	    first_entry = src;
	else			/* dst is not the first node in the list */
	    dst->prev->next = src;
	src->next = dst;
	dst->prev = src;

	/* the entry is being moved down, */
	/* so insert src just after the node dst points to */
    } else {
	src->next = dst->next;
	if (dst->next)		/* if dst is not the last node in the list */
	    dst->next->prev = src;
	dst->next = src;
	src->prev = dst;
    }
}

void remove_entry(struct menu *menu, bool * playlist_changed)
{
    if (!menu->entries) {
	msg("There are no entries.");
	return;
    }
    actually_remove_entry(menu->current);
    *playlist_changed = TRUE;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    menu->entries--;
    if (menu->entries > 0) { /* a check to avoid menu->current becoming -1 */
	if (menu->current > menu->entries - 1)
	    menu->current--;
    }
}

void actually_remove_entry(int index)
{
    int i = 0;
    bool only_one_node_in_list = FALSE;
    struct playlist_entry *p = first_entry;

    assert(first_entry != NULL);

    while (i++ < index)
	p = p->next;

    if (p->prev && p->next) {
	p->prev->next = p->next;
	p->next->prev = p->prev;
    } else if (p->prev)
	p->prev->next = NULL;
    else if (p->next) {		/* this only happens if index is 0 */
	p->next->prev = NULL;
	first_entry = first_entry->next;
    } else			/* else p points to first_entry and that's the only entry in the list */
	only_one_node_in_list = TRUE;

    safe_free(p->path);
    safe_free(p->song);
    safe_free(p);

    if (only_one_node_in_list == TRUE)
	first_entry = NULL;
}

void interactive_save_playlist(struct menu *menu, char **playlist, bool * playlist_changed)
{
    char path[PATH_MAX + 1];
    int ret = -1;

       if (*playlist_changed == FALSE) {
	 msg("No changes need to be saved."); 
	 return;
       }

       if (*playlist)		/* if we already have a playlist open */
	ret = save_playlist(*playlist, OVERWRITE);
    else {
      snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");
	if (enter_path(path, sizeof(path), REG_FILE | HALF_FILLED, "Save playlist to: ")) {
	    ret = save_playlist(path, ASK_IF_EXISTS);
	    if (ret == 0) {
		*playlist = safe_strdup(path);
		/* refresh the playlist name in the status bar */
		TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);		
	    }
	}
    }

    if (ret != 0)
	return;

    /* this will only be displayed if the first 'if' was true and */
    /* save_playlist() returned 0 (indicating success) */
    msg("Wrote playlist to %s.", *playlist);
    *playlist_changed = FALSE;
}

void interactive_save_playlist_as(struct menu *menu, char **playlist, bool * playlist_changed)
{
    char path[PATH_MAX + 1];

      snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");
    if (enter_path(path, sizeof(path), REG_FILE | HALF_FILLED, "Save playlist to: ")) {
	if (save_playlist(path, ASK_IF_EXISTS) != 0)
	    return;		/* saving failed */
    } else			/* aborted */
	return;

    free(*playlist);
    *playlist = safe_strdup(path);
    msg("Wrote playlist to %s.", *playlist);
    *playlist_changed = FALSE;
    TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
}

/*
 * Save the linked list starting with first_entry to playlist.
 * Return 0 on success, -1 on error.
 */
int save_playlist(const char *playlist, bool ask_if_exist)
{
    struct playlist_entry *p;
    char buf[SHORT_STRING];
    FILE *fp;

    if (ask_if_exist) {
	if (access(playlist, F_OK) == 0) {
	    snprintf(buf, sizeof(buf), "%s exists already, overwrite?", playlist);
	    if (!yes_or_no(0, buf))
		return -1;
	}
    }
    fp = fopen(playlist, "w");
    if (!fp) {
	perror_msg("Can't write playlist to %s", playlist);
	return -1;
    }
    if (first_entry) {		/* if there are no entries, the written playlist will be empty */
	for (p = first_entry; p; p = p->next)
	    fprintf(fp, "%s%s\n", p->path, p->toggled ? " *" : "");
    }
    (void) fclose(fp);

    return 0;
}

/*
 * Ask user for a file to write the song names in the playlist to, and then
 * write the song names to a file.
 */
void interactive_write_song_names_to_file(struct menu *menu)
{
    char path[PATH_MAX + 1];

    snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");
    if (enter_path(path, sizeof(path), REG_FILE | HALF_FILLED,
		   "Write song names to: ")) {
	/* if operation not fullfilled */
	if (write_song_names_to_file(path, ASK_IF_EXISTS) != 0) {
	    return;
	}
    } else			/* user aborted */
	return;

    msg("Wrote song names to %s.", path);
}

/*
 * Write the song names in the playlist (the song member of the linked list
 * starting with first_entry) to file, each one separated by newline.
 * Return 0 on success, -1 on error.
 */
static int write_song_names_to_file(const char *file, bool ask_if_exist)
{
    struct playlist_entry *p;
    char buf[STRING];
    FILE *fp;

    if (ask_if_exist) {
	if (access(file, F_OK) == 0) {
	    snprintf(buf, sizeof(buf), "%s exists already, overwrite?", file);
	    if (!yes_or_no(0, buf))
		return -1;
	}
    }
    fp = fopen(file, "w");
    if (!fp) {
	perror_msg("Can't write to %s", file);
	return -1;
    }
    if (first_entry) {
	for (p = first_entry; p; p = p->next)
	    fprintf(fp, "%s\n", p->song);
    }	/* else there are no entries and the written file will be empty */
    (void) fclose(fp);

    return 0;
}

/*
 * Do a regular expression search through the `song' or `path' member
 * (depending on the `OPT_SHOW_PATH' option), of each playlist entry until a
 * match is found or all entries have been searched. If the former happens,
 * update `menu' so the matching entry will be highlighted on next redraw.
 */
void playlist_search(struct menu *menu, int op)
{
    /* current playlist entry being compared to the regexp */
    struct playlist_entry *cur;
    /* index position of current playlist entry being compared to the regexp */
    int cur_index_pos;
    char search_str[32];     /* search string */
    bool first_time_through; /* used by the search loop */
    regex_t preg;            /* compiled regexp structure, must be regfree()d*/
    int errcode;             /* error code to check for regcomp() error */

    /* make `cur' point to the current entry in the playlist and
       `cur_index_pos' contain its position/index number */
    cur = first_entry;
    cur_index_pos = 0;
    while (cur_index_pos++ < menu->current)
	cur = cur->next;
    cur_index_pos--;

    if (op == CMD_REPEAT_LAST_SEARCH || op == CMD_REPEAT_LAST_SEARCH_OPPOSITE) {
        if (menu->prev_main_search_op == NOP) {
	    msg("No previous searches.");
	    return;
	}
	strfcpy(search_str, menu->prev_search_str, sizeof(search_str));
	if (op == CMD_REPEAT_LAST_SEARCH_OPPOSITE) {
	    /* opposite direction of previous main search operation */
	    if (menu->prev_main_search_op == CMD_REGEXP_SEARCH_BACKWARD)
		op = CMD_REGEXP_SEARCH_FORWARD;
	    else
		op = CMD_REGEXP_SEARCH_BACKWARD;
	} else { /* op == CMD_REPEAT_LAST_SEARCH */
    	    op = menu->prev_main_search_op;
	}
    } else if (op == CMD_REGEXP_SEARCH_BACKWARD || op == CMD_REGEXP_SEARCH_FORWARD) {
	if (enter_string(search_str, sizeof(search_str),
			 EMPTY, NO_BROWSE, "%s", op == CMD_REGEXP_SEARCH_FORWARD ?
			 "Search for: " :
			 "Reverse search for: ") == NULL) {
	    return;
	}
	strfcpy(menu->prev_search_str, search_str,
		sizeof(menu->prev_search_str));
	menu->prev_main_search_op = op;
    } else /* unknown search operation */
	assert(0);

    /* start search from the next or previous line */
    if (op == CMD_REGEXP_SEARCH_FORWARD) {
	cur = cur->next;
	cur_index_pos++;
    } else if (op == CMD_REGEXP_SEARCH_BACKWARD) {
	cur = cur->prev;
	cur_index_pos--;
    }

    /* compile `search_str' into a regular expression */
    if (is_boolean_option_true(OPT_IGNORE_CASE))
	errcode = regcomp(&preg, search_str,
			  REG_NOSUB | REG_ICASE | REG_EXTENDED);
    else /* do not ignore case */
	errcode = regcomp(&preg, search_str, REG_NOSUB | REG_EXTENDED);

    if (errcode) {
	display_regexp_error(errcode, &preg);
	return;
    }

    first_time_through = TRUE;
 loop_again:
    while (cur) {

	/* does `cur->song' or `cur->path' match the regexp `preg'? */
	if (is_boolean_option_true(OPT_SHOW_PATH)) {
	    if (regexec(&preg, cur->path, -1, NULL, 0) == 0)
		goto found;
	} else {
	    if (regexec(&preg, cur->song, -1, NULL, 0) == 0)
		goto found;
	}

	if (op == CMD_REGEXP_SEARCH_FORWARD) {
	    if (cur->next)
		cur_index_pos++;
	    cur = cur->next;
	} else if (op == CMD_REGEXP_SEARCH_BACKWARD) {
	    if (cur->prev)
		cur_index_pos--;
	    cur = cur->prev;
	} else
	    assert(0);
    }

    /* if we're on the first time through, we wrap the search */
    if (first_time_through == TRUE) {
	first_time_through = FALSE;
	cur = first_entry;
	cur_index_pos = 0;
	if (op == CMD_REGEXP_SEARCH_BACKWARD) {
	    while (cur->next) {
		cur_index_pos++;
		cur = cur->next;
	    }
	} /* else op == CMD_REGEXP_SEARCH_FORWARD, so don't do anything since
	     `cur' points to the first entry and `cur_index_pos' is 0 */

	goto loop_again;
    }

    msg("Not found.");
    regfree(&preg); /* `preg' contains pointers to allocated memory blocks */
    return;

 found:
    /* if search was wrapped */
    if (first_time_through == FALSE) {
	msg("Search hit %s.",
		    op == CMD_REGEXP_SEARCH_FORWARD ?
		    "bottom, continuing at top" :	/* forward search */
		    "top, continuing at bottom");	/* backward search */
    }
    if (cur_index_pos != menu->current) {
	menu->current = cur_index_pos;
	TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
    }
    regfree(&preg); /* `preg' contains pointers to allocated memory blocks */
}

void toggle_all_entries(bool yes)
{
    struct playlist_entry *p = first_entry;

    while (p) {
	p->toggled = yes;	/* yes is either 1 or 0 */
	p = p->next;
    }
}

void toggle_entry(int index)
{
    struct playlist_entry *p = first_entry;
    int i = 0;

    while (i++ < index)
	p = p->next;

    p->toggled = !p->toggled;
}

void sort_playlist(void)
{
    struct playlist_entry *fu;	/* the first unsorted node to be inserted */
    struct playlist_entry *ls;	/* the last sorted node (tail of sorted sublist) */
    struct playlist_entry *cur, *trailing;

    if (!first_entry)
	return;

    ls = first_entry;		/* An empty list is already sorted. */
    while (ls->next) {
	fu = ls->next;		/* Remember first unsorted node.    */

	if (is_boolean_option_true(OPT_SHOW_PATH) ?
	    LT(fu->path, first_entry->path) :
	    LT(fu->song, first_entry->song)) {
	    ls->next = fu->next;
	    fu->next = first_entry;
	    fu->prev = NULL;
	    first_entry->prev = fu;
	    first_entry = fu;	/* Insert first unsorted at the head of sorted list. */
	} else {		/* Search the sorted sublist. */
	    trailing = first_entry;
	    for (cur = trailing->next;
		 is_boolean_option_true(OPT_SHOW_PATH) ? GT(fu->path, cur->path) : GT(fu->song, cur->song);
		 cur = cur->next)
		trailing = cur;

	    if (fu == cur)
		ls = fu;
	    else {		/* First unsorted node now belongs between trailing and cur. */
		ls->next = fu->next;
		if (fu->next)
		    fu->next->prev = ls;
		fu->next = cur;
		cur->prev = fu;
		fu->prev = trailing;
		trailing->next = fu;
	    }
	}
    }
}

void reverse_sort_playlist(void)
{
    struct playlist_entry *fu;	/* the first unsorted node to be inserted */
    struct playlist_entry *ls;	/* the last sorted node (tail of sorted sublist) */
    struct playlist_entry *cur, *trailing;

    if (!first_entry)
	return;

    ls = first_entry;		/* An empty list is already sorted. */
    while (ls->next) {
	fu = ls->next;		/* Remember first unsorted node.    */

	if (is_boolean_option_true(OPT_SHOW_PATH) ?
	    GT(fu->path, first_entry->path) :
	    GT(fu->song, first_entry->song)) {
	    ls->next = fu->next;
	    fu->next = first_entry;
	    fu->prev = NULL;
	    first_entry->prev = fu;
	    first_entry = fu;	/* Insert first unsorted at the head of sorted list. */
	} else {		/* Search the sorted sublist. */
	    trailing = first_entry;
	    for (cur = trailing->next;
		 is_boolean_option_true(OPT_SHOW_PATH) ? LT(fu->path, cur->path) : LT(fu->song, cur->song);
		 cur = cur->next)
		trailing = cur;

	    if (fu == cur)
		ls = fu;
	    else {		/* First unsorted node now belongs between trailing and cur. */
		ls->next = fu->next;
		if (fu->next)
		    fu->next->prev = ls;
		fu->next = cur;
		cur->prev = fu;
		fu->prev = trailing;
		trailing->next = fu;
	    }
	}
    }
}
